#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# BrEPScmd performes searches in the BrEPS enzyme sequence patten
# database created with BrEPSdb.
# Copyright (C) 2017 Christian-Alexander Dudek <c.dudek@tu-braunschwig.de>,
# Technische Universität Braunschweig
# Department of Bioinformatics and Biochemistry
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
``brepscmd.py``
---------------
The BrEPS command line tool (BrEPScmd) can be used to search a local copy
of the database. The tool needs either Python2 or Python 3 to run.
For a quick help, please see the `online documentation`_.

This documentation lists all functions used by BrEPScmd and can be used
as guide for tool development.

For MySQL usage, `mysqlclient`_ is recommended, as it can be used with
Python2 and Python3. For Python2 `MySQLdb`_ is suitable.

For best performance, Python2 with `Googles re2 module`_ is recommended!

.. _Googles re2 module: https://github.com/google/re2
.. _online documentation: http://breps.tu-bs.de/help/brepscmd.html
.. _mysqlclient: https://github.com/PyMySQL/mysqlclient-python
.. _MySQLdb: https://sourceforge.net/projects/mysql-python/
"""
import os
import sys
import time
import argparse
import multiprocessing as mp
import signal


try:
    import sqlite3 as sqlt
except ImportError:
    print("SQLite not installed!")

try:
    import MySQLdb as mysql
except ImportError:
    print("MySQLdb not installed!")

try:
    import re2 as re
except ImportError:
    import re
    if sys.version_info < (3,):
        print("Using re module, install re2 (https://github.com/google/re2) "
              "for better performance or use Python3 instead.")

dctDeli = {"comma": ",",
           "semi": ";",
           "tab": "\t",
           "space": " "}

dctOneLetterCode = {"ACC": "T", "ATG": "M", "AAG": "K", "AAA": "K", "ATC": "I",
                    "AAC": "N", "ATA": "I", "AGG": "R", "CCT": "P", "CTC": "L",
                    "AGC": "S", "ACA": "T", "CTT": "L", "CAT": "H", "AAT": "N",
                    "ATT": "I", "CTG": "L", "CTA": "L", "ACT": "T", "CAC": "H",
                    "ACG": "T", "CAA": "Q", "AGT": "S", "CAG": "Q", "CCG": "P",
                    "CCC": "P", "TAT": "Y", "GGT": "G", "TGT": "C", "CGA": "R",
                    "CCA": "P", "TCT": "S", "GAT": "D", "CGG": "R", "TTT": "F",
                    "TGC": "C", "GGG": "G", "TAG": "-", "GGA": "G", "TAA": "-",
                    "GGC": "G", "TAC": "Y", "TTC": "F", "TCG": "S", "TTA": "L",
                    "AGA": "R", "GAC": "D", "TCC": "S", "GAA": "E", "TCA": "S",
                    "GCA": "A", "GTA": "V", "GCC": "A", "GTC": "V", "GCG": "A",
                    "GTG": "V", "GAG": "E", "GTT": "V", "GCT": "A", "TGA": "-",
                    "TTG": "L", "CGT": "R", "TGG": "W", "CGC": "R"}

strDNA = "ATGC"
strAA = "TMKINRPSLHQYGCDFEAVW"

strRep = r"\((|\d+,)?(\d+)\)-?"
strSel = r"-?\[([ARNDCQEGHILKMFPSTWYVBZX]*)\]-?"
strAA = r"-?([ARNDCQEGHILKMFPSTWYVBZX])-?"
strAny = r"-?([x])-?"
strAcc = (r"([OPQ][0-9][A-Z0-9]{3}[0-9]|[A-NR-Z][0-9]"
          "([A-Z][A-Z0-9]{2}[0-9]){1,2})")
strEc = r"[1-6]\.[0-9\-]{1,2}\.[0-9\-]{1,2}\.[0-9\-]{1,3}"


# Timeout exception for the verification timeout.
class TimeoutExc(Exception):
    pass


def timeout(intSecs, *args):
    """
    Timeout decorator for the pattern verification.

    Complex regular expression search may run for a long time causing
    the verification to get stuck. The decorator adds a signal that
    kills a verification that runs longer than ``intSecs`` seconds.

    Args:
        intSecs (integer): Seconds after which the timeout is fired.
    """
    def decorate(f):
        def handler(signum, frame):
            raise TimeoutExc("Timeout!")

        def new_f(*args):
            signal.signal(signal.SIGALRM, handler)
            signal.alarm(intSecs)
            return f(*args)
            signal.alarm(0)

        new_f.__name__ = f.__name__
        return new_f
    return decorate


def parse_fasta(strFasta):
    """
    Parses the contents of a FASTA file.

    Args:
        strFasta (str): Contents of a FASTA file.

    Returns:
        dict: A dictionary with sequence identifier as key and sequence
        as value.
    """
    dctResult = {}
    strSeq = ""
    strId = None
    for line in strFasta.split("\n"):
        if line != "" and line[0] == ">":
            if strId is not None:
                dctResult[strId] = strSeq
                strSeq = ""
            strId = line[1:81]
        else:
            strSeq += line
    if strId is not None and strSeq:
        dctResult[strId] = strSeq
    return dctResult


def ps2re(strPS):
    """
    Converts PROSITE pattern to regular expression.

    Args:
        strPS (str): Pattern in PROSITE format.

    Returns:
        str: Pattern in regular expression format.

    """
    strRe = re.sub("-?\[([ARNDCQEGHILKMFPSTWYVBZX]*)\]-?", "[\\1]", strPS)
    strRe = re.sub("\((|\d+,)?(\d+)\)-?", "{\\1\\2}", strRe)
    strRe = re.sub("-?([ARNDCQEGHILKMFPSTWYVBZX])-?", "\\1", strRe)
    strRe = re.sub("-?([x])-?", ".", strRe)
    return strRe


def re2ps(strRe):
    """
    Converts regular expression pattern to PROSITE.

    Args:
        strRe (str): Pattern in regular expression format.

    Returns:
        str: Pattern in PROSITE format.
    """
    strPS = re.sub("(\[[A-Z]*\])", "-\\1-", strRe)
    strPS = re.sub("\.\{([0-9]*)\}", "-x(\\1)-", strPS)
    strPS = re.sub("([A-Z])\{([0-9]*)\}", "-\\1(\\2)-", strPS)
    strPS = re.sub("\.\{([0-9]*,[0-9]*)\}", "-x(\\1)-", strPS)
    strPS = re.sub("\.", "-x-", strPS)
    lstPS = [x for x in strPS.split("-") if x]
    lstPS = ["-".join(x)
             if ("[" not in x and "(" not in x) else x for x in lstPS]
    return "-".join(lstPS)


def dna2aa(strDNA):
    """
    Converts a DNA sequence to an amino acid sequence.

    Args:
        strDNA (str): Nucleic acid sequence.

    Returns:
        str: The translated amino acid sequence.
    """
    lstAA = [dctOneLetterCode[strDNA[i:i+3]]
             for i in range(0, len(strDNA), 3)
             if len(strDNA[i:i+3]) == 3 and
             strDNA[i:i+3] in dctOneLetterCode]
    return "".join(lstAA)


def complexity(strPS):
    """
    Calculates the complexity of a Prosite pattern.

    The pattern complexity is calculated for every aminoacid by the
    following rules:

        - A conserved aminoacid has complexity 1/20
        - Set of n aminoacids has complexity of n/20
        - Wildcard positions (x) have complexity of 1

     The sum of all complexities will be divided by the maximum length
     of the pattern.

    Args:
        strPS (str): Pattern in PROSITE format.

    Returns:
        Complexity as float ranged from 1.0 (complex) to 0.0 (not complex).
    """
    lstPS = strPS.split("-")
    complexity = 0.0
    min_length = 0
    max_length = 0
    for pos in lstPS:
        if "[" in pos:
            complexity += (len(pos)-2)/20.0
            min_length += 1
            max_length += 1
        elif "x" in pos and "," in pos:
            n = pos.replace("x(", "").replace(")", "").split(",")
            mn = int(n[0])
            mx = int(n[1])
            complexity += mx*1.0
            min_length += mn
            max_length += mx
        elif "x" in pos and "(" in pos:
            num = int(pos.replace("x(", "").replace(")", ""))
            complexity += num*1.0
            min_length += num
            max_length += num
        elif "x" == pos:
            complexity += 1.0
            min_length += 1
            max_length += 1
        else:
            complexity += 1/20.0
            min_length += 1
            max_length += 1
    return complexity/max_length


def get_patterns(dbCur, intMaxLen=7000, fltMaxEvalue=1.0, lstExclude=[]):
    """
    Gets all sequences with a maximum length of ``<intMaxLen>``.

    Args:
        dbCur (Dbase): Database cursor instance of :class:`.DBase`.
        intMaxLen (int): The maximum length of a pattern
            (Default: 7000, basically all patterns).
        fltMaxEvalue (float): The maximum E-value of a pattern
            (Default: 1.0, basically all patterns).
        lstExclude (list): A list of pattern types to exclude
            (Default: [], exclude no pattern types).

    Returns:
        list: List of lists with all patterns for a breps id.
    """
    dctRes = {}
    lstType = [x for x in ["std", "ext"] if x not in lstExclude]
    # Get all patterns based on minimum length
    dbCur.execute("SELECT p.pat_id,p.type,p.breps_id,p.pat_re,p.pat_sub,"
                  "p.min_len FROM patterns p "
                  "LEFT JOIN annotation a ON p.breps_id=a.breps_id "
                  "WHERE p.min_len <= %s AND a.evalue < %s",
                  (intMaxLen, fltMaxEvalue))
    # Remove patterns if type not needed
    lstRes = []
    for p in dbCur.fetchall():
        if p[1] in lstType:
            lstRes.append(p)
    # Sort all pattern types based on breps_id into a dictionary
    for p in lstRes:
        if p[2] not in dctRes:
            dctRes[p[2]] = []
        dctRes[p[2]].append(p)
    # Convert dictionary into list. Every list item is a list with
    # patterns associated with one breps_id.
    lstRes = [dctRes[x] for x in dctRes]
    return lstRes


@timeout(1)
def re_match(reSub, rePat, strSeq):
    if ((bool(re.search(reSub, strSeq)) and
         bool(re.search(rePat, strSeq)))):
        return True
    else:
        return False


def search(lstPatterns):
    """
    Performs the BrEPS search.

    Sequences and patterns are given as dictionries in a tuple for
    multirocessing. The search first checks the standard pattern checks
    and only checks extended patterns if the standard pattern doesn't
    hit. The dictionry of sequences resides outside of this function.
    This is neccessary for multithreading.

    Args:
        lstPatterns (list): List with patterns for one BrEPS id to
            check.

    Returns:
        dict: A dictionry with sequence id as key and a list of
        pattern ids for every hit.
    """
    lstHits = []
    for intSeqid in dctSeqs:
        strSeq = dctSeqs[intSeqid]

        # We don't want to check every pattern for a breps_id. If a
        # standard pattern hits, extended and subpatterns will also hit.
        # So we sort patterns by type: std > ext.
        lstSort = ["std", "ext"]
        lstPatterns = sorted(lstPatterns, key=lambda x: lstSort.index(x[1]))
        tplHit = None
        for pat_id, typ, breps_id, pat_re, pat_sub, min_len in lstPatterns:
            if (min_len <= len(strSeq)):
                blnMatch = False
                try:
                    blnMatch = re_match(pat_sub, pat_re, strSeq)
                except TimeoutExc:
                    pass

                if (blnMatch):
                    tplHit = (pat_id, intSeqid)
                    break

        if tplHit is not None:
            lstHits.append(tplHit)
            tplHit = None
    if lstHits:
        return lstHits
    else:
        return []


def prepare_data(lstHits, dbCur, strOrder="evalue"):
    """
    Gets the annotations for a list of BrEPS hits.

    Pattern and annotation data is gathered and prepared for output.

    Args:
        dctHits (dict): Dictionary with sequence ids as keys and list
            of pattern ids as values from :func:`.search`.
        dbCur (Dbase): Database cursor instance of :class:`.DBase`.
        strOrder (str): Database column name to sort the results.

    Returns:
        None.

    Raises:
        None.
    """
    lstResults = []
    for h in lstHits:
        if h is not None:
            lstResults.extend(h)
    if not lstResults:
        return []
    setPatIds = {r[0] for r in lstResults}
    # Get all pattern data found while searching
    dbCur.execute("SELECT * FROM patterns WHERE pat_id IN ({})".format(
        ",".join([str(x) for x in setPatIds])))

    dctPats = {r[0]: dict(
        zip([i[0].replace("_", "") for i in dbCur.description], r))
               for r in dbCur.fetchall()}

    # Get all annotations associated with the search results
    setBrepsIds = {dctPats[x]["brepsid"] for x in dctPats}

    dbCur.execute("SELECT * FROM annotation WHERE breps_id IN ({})".format(
        ",".join([str(x) for x in setBrepsIds])))
    dctAnno = {r[0]: dict(
        zip([i[0].replace("_", "") for i in dbCur.description], r))
               for r in dbCur.fetchall()}
    # Iter results
    lstRes = []
    for pat_id, strSeqId in lstResults:
        breps_id = dctPats[pat_id]["brepsid"]

        # Combine pattern data with annotation
        dctLine = dctPats[pat_id].copy()
        dctLine.update(dctAnno[breps_id])

        # Extend data
        dctLine["seqid"] = strSeqId
        dctLine["rawid"] = dctLine["brepsid"]
        dctLine["brepsid"] = "bp{:0>6}".format(dctLine["rawid"])
        # Only calculate position and length if needed
        if "hitpos" in lstColumns or "hitlen" in lstColumns:
            m = re.search(dctLine["patre"], dctSeqs[strSeqId])
            dctLine["hitpos"] = m.start()
            dctLine["hitlen"] = m.end() - m.start()
        lstRes.append(dctLine)
    if strOrder in lstRes[0]:
        lstRes = sorted(lstRes, key=lambda x: x[strOrder])
    return lstRes


class Dbase(object):
    """
    Database class for both MySQL and SQLite usage.

    On creation the ``strType`` can be selected from ``MySQL`` and
    ``SQLite`` to select a database type. All functions are created just
    like the provided by MySQLdb or sqlite3. MySQL placeholders ``%s``
    are automatically replaced for SQLite. Before every execution the
    connection is checked to avoid ``mysql.OperationalError``
    exceptions. The returned instance can be directly used for
    execution. Also the instance needs to be closed properly.

    **Only non-standard methods are documented!**

    Attributes:
        strType (str): Type of database connection (mysql or sqlite).
        strDB (str): Database name for MySQL, path to file for SQLite.
        strHost (str): Host for the MySQL connection.
        strUser (str): Username for the MySQL connection.
        strPass (str): Password for the MySQL connection.

    """
    def __init__(self, strType, strDB, strHost=None,
                 strUser=None, strPass=None):
        super(Dbase, self).__init__()
        self.strType = strType
        self.strDB = strDB
        self.strHost = strHost
        self.strUser = strUser
        self.strPass = strPass
        self.connect()
        self.description = None

    def connect(self):
        if self.strType == "mysql":
            self.db = mysql.connect(self.strHost,
                                    self.strUser,
                                    self.strPass,
                                    self.strDB)
            self.s = "%s"
        elif self.strType == "sqlite":
            self.db = sqlt.connect(self.strDB)
            self.s = "?"
        self.cur = self.db.cursor()

    def execute(self, strQry, lstData=[], commit=True):
        try:
            res = self.cur.execute(strQry.replace("%s", self.s), lstData)
        except mysql.OperationalError as e:
            if e[0] == 2006:
                self.db.close()
                self.connect()
                res = self.execute(strQry.replace("%s", self.s), lstData)
            else:
                raise e
        self.description = self.cur.description
        if commit:
            self.db.commit()
        return res

    def executemany(self, strQry, lstData=[], commit=True):
        # We want to split huge queries into smaller ones...
        # Large queries can cause Error 2006, (MySQl server has gone
        # away)!

        if lstData:
            lstChunks = [lstData[i:i+200]
                         for i in xrange(0, len(lstData), 200)]
            for chunk in lstChunks:
                try:
                    res = self.cur.executemany(strQry.replace("%s", self.s),
                                               chunk)
                except mysql.OperationalError as e:
                    if e[0] == 2006:
                        self.connect()
                        res = self.cur.executemany(strQry.replace("%s",
                                                   self.s), chunk)
                    else:
                        raise e
                self.description = self.cur.description
            if commit:
                self.commit()

            return res
        else:
            return None

    def load_sql(self, strFile):
        os.system(
            "mysql -h %s -u %s, -p%s %s < %s".format(
                self.strHost, self.strUser, self.strPass, self.strDB, strFile
                )
            )

    def fetchall(self):
        return self.cur.fetchall()

    def fetchone(self):
        return self.cur.fetchone()

    def commit(self):
        self.db.commit()

    def close(self):
        self.cur.close()
        self.db.close()


def print_results(lstLines, lstColumns, clsOut, strDelimiter, strFormat="csv"):
    """
    Prints the data results to terminal or into a file.

    Fields containing a delimiter sign will be surrounded by double
    quotes.

    ..todo: Add alignment output format!

    Args:
        lstLines (list): List of prepared data to display, one for each
            pattern hit.
        lstColumns (list): Columns to display.
        clsOut (instance): Class instance to write to. Needs a
            ``write()`` method!.
        strDelimiter (str): CSV field delimiter.
        strFormat (str): Output format, currently only CSV!
    """
    strOut = strDelimiter.join(["{%s}" % x for x in lstColumns])
    strOut += "\n"
    clsOut.write(strDelimiter.join(lstColumns) + "\n")

    for dctLine in lstLines:
        # Escape data with strDelimiter with "
        for x in dctLine:
            if (isinstance(dctLine[x], (str)) and
                    strDelimiter in dctLine[x]):
                dctLine[x] = "\"{}\"".format(dctLine[x])
        clsOut.write(strOut.format(**dctLine))

# All supported output columns. Basically these are the database field
# names without the underscore plus additional columns.
dctColumns = {"brepsid": "BrEPS id (bpxxxxxx)",
              "rawid": "The raw id of the BrEPS entry (without bp)",
              "patid": "The unique pattern id",
              "type": "Type of pattern <standrad, extended, sub>",
              "propec": "Proposed EC number based on associated EC numbers",
              "propfunc": "Proposed function based on IUBMB nomenclature of "
                          "proposed EC number",
              "seqid": "Sequence id got from fasta file",
              "ecs": "ECs separated by & , groups by separated ,",
              "evalue": "E-value of the pattern node",
              "numecs": "Integer of distinct EC groups",
              "comp": "Pattern complexity",
              "hitpos": "Relative position of pattern hit",
              "hitlen": "Length of the pattern hit",
              "patps": "Pattern in PROSITE format",
              "patre": "Pattern in regex format",
              "minlen": "Minimum length of the pattern",
              "maxlen": "Maximum length of the pattern"}


class ColumnsAction(argparse.Action):
    """
    Argparser action to display usable columns for CSV output.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        intMaxlen = max([len(x) for x in dctColumns.keys()])
        strRepr = "{:<%i} {}" % (intMaxlen + 3)
        print("You can specify a custom output format for the csv output.\n"
              "Use '-c id1 id2 ... idN' to select data rows.")
        print("Supported ids and their meaning:")

        print(strRepr.format("ID", "MEANING"))
        for key in dctColumns:
            print(strRepr.format(key, dctColumns[key]))
        exit()


class LicenseAction(argparse.Action):
    """
    Argparser action to display Disclaimer of Warranty.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        print("The program can be redistributed under GPL3 conditions. "
              " Please see the `LICENSE' file or "
              "https://opensource.org/licenses/GPL-3.0 for detailed "
              "license information.")
        exit()


class WarrantyAction(argparse.Action):
    """
    Argparser action to display Disclaimer of Warranty.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        print("THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED "
              "BY APPLICABLE LAW. EXCEPT WHEN OTHERWISE STATED IN WRITING THE "
              "COPYRIGHT HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "
              "'AS IS' WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR "
              "IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES "
              "OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE "
              "ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM "
              "IS WITH YOU. SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME "
              "THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.")
        exit()

if __name__ == "__main__":
    print("BrEPScmd Copyright (C) 2017 Christian-Alexander Dudek,\n"
          "Technische Universität Braunschweig\n"
          "Department of Bioinformatics and Biochemistry\n\n"
          "This program comes with ABSOLUTELY NO WARRANTY; for details type "
          "`python brepscmd.py -w'. This is free software, and you are "
          "welcome to redistribute it under certain conditions; type "
          "`python brepscmd.py -l' for license details.\n")
    # The Argparser handles the user input via command line.
    # Options can be given as command line argument or JSON config file.
    prsr = argparse.ArgumentParser(
        epilog="One of the argument -i is mandatory! By default SQLite "
               "with file breps_db.db will be used.")

    # Version -v = Shows BrEPScmd and BrEPSdb version
    prsr.add_argument(
        "-v", "--version", action="version",
        version="%(prog)s version 2.0; Database version 2016.1")
    prsr.add_argument(
        "--columns", action=ColumnsAction, nargs=0,
        help="show detailed list of columns and exit")
    # Input file -i = Path to file in fasta format or plain sequence
    prsr.add_argument(
        "-i", metavar="path/sequence",
        help="input file in FASTA format or plain sequence", default=None)
    # Output file -o = Path to file for storing the results
    prsr.add_argument(
        "-o", metavar="path/to/file.ext", type=argparse.FileType(mode="w"),
        default=sys.stdout,  help="output filename, default: Stdout")
    # Number of threads -t
    prsr.add_argument(
        "-t", type=int, metavar="threads", default=1,
        help="number of threads used for searching")
    # Delimiter (for csv) -l = ,
    prsr.add_argument(
        "-d", default="comma", choices=["comma", "semi", "tab"],
        help="delimiter for csv output")
    # Data columns -c
    prsr.add_argument(
        "-c", metavar="id", nargs="+",
        help="define the columns of the csv output. (Use --columns to "
             "get a list of possible values)")
    # Exclude pattern types from search
    prsr.add_argument(
        "-e", metavar="type", nargs="+",
        help="exclude pattern types (std, ext, sub) from search")
    # E-value cut-off
    prsr.add_argument(
        "-eva", metavar="E-value", type=float, default=1.0,
        help="pattern E-value cut-off. Only patterns with lower E-values "
             "will be checked")
    # Database arguments, -db can be the MySQL database name or a
    # SQLite database file. MySQL will be used if all arguments are
    # given. SQLite will use the given database or breps_db.db if no
    # database file is given.
    prsr.add_argument(
        "-db", metavar="database",
        help="MySQL database name or path to SQLite file")
    prsr.add_argument(
        "-host", metavar="host",
        help="MySQL database host")
    prsr.add_argument(
        "-user", metavar="username",
        help="MySQL database user")
    prsr.add_argument(
        "-pw", metavar="password",
        help="MySQL database password")
    prsr.add_argument("-w", "--warranty", action=WarrantyAction, nargs=0,
                      help="show disclaimer of warranty and exit")
    prsr.add_argument("-l", "--license", action=LicenseAction, nargs=0,
                      help="show license information and exit")

    args = prsr.parse_args()

    # Prepare arguments
    # Check if sequence or file is given
    if args.i is None:
        sys.stdout.write("No input sequence given.\nAdd -i file/ "
                         "sequence to your command\n")
        sys.exit()
    # Given input is a file
    elif os.path.isfile(args.i):
        fh = open(args.i)
        dctSeqs = parse_fasta(fh.read())
        fh.close()
    # Input is a plain sequence
    else:
        dctSeqs = {}
        if all(c in strDNA for c in args.i):
            # convert dna to aa
            strSeq = dna2aa(args.i)
        elif not all(c in strAA for c in args.i):
            # there is no aa seq
            sys.stdout.write("No amino acid or nucleic acid sequence given\n")
            sys.exit()
        else:
            strSeq = args.i
        dctSeqs["seq1"] = strSeq
    # Data columns to show in csv results.
    lstColumns = []
    if args.c is not None:
        lstColumns = [x for x in args.c if x in dctColumns.keys()]
    if not lstColumns:
        lstColumns = ["seqid", "brepsid", "type", "propec",
                      "propfunc", "evalue"]
    # A specific pattern type can be excluded for search.
    if args.e is not None:
        lstExclude = args.e
    else:
        lstExclude = []
    # Connect to database, use MySQL or SQLite based on user input.
    if (args.db is not None and
            args.host is not None and
            args.user is not None and
            args.pw is not None):
        cur = Dbase("mysql", args.db, args.host, args.user, args.pw)
    elif (args.db is not None and os.path.isfile(args.db)):
            cur = Dbase("sqlite", args.db)
    elif os.path.isfile("brepsdb.db"):
        cur = Dbase("sqlite", "brepsdb.db")
    else:
        print("No database found. Specify MySQL/SQLite data or provide "
              "brepsdb.db SQLite file in this directory!")
        sys.exit()

    # When all arguments are valid and a database is present, the search
    # starts.
    # Get the maximum length of all patterns
    intMaxLen = max([len(dctSeqs[s]) for s in dctSeqs])

    # List of all patterns to check
    lstPatterns = get_patterns(cur, intMaxLen, args.eva, lstExclude)

    # Use multiprocessing to search the database.
    t0 = time.time()
    pool = mp.Pool(processes=args.t)
    lstHits = pool.map(search, lstPatterns)

    # Get allannotation and pattern data for the hits.
    lstData = prepare_data(lstHits, cur)

    # Print the results either to terminal or into file.
    print_results(lstData, lstColumns, args.o, dctDeli[args.d])

    t1 = time.time()
    print("Done after {:.2f}s\n".format((t1-t0)))
